from ..structs import Generate, User
from ..turbo import turbo

__all__ = [
    "summarize_bot",
]

# Templates
SUMMARIZE_TEMPLATE = """
Summarize the following {{text_type}}:

{{text}}
""".strip()


@turbo(model="gpt-3.5-turbo", temperature=0.2)
async def summarize_bot(text: str, text_type: str = "text"):
    yield User(
        template=SUMMARIZE_TEMPLATE,
        variables=dict(text=text, text_type=text_type),
    )

    yield Generate()



from typing import (
    Any,
    Callable,
)

import openai

from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)


__all__ = [
    "create_retry_decorator",
    "with_retries",
]


# Retries
def create_retry_decorator(
    min_seconds: int = 4,
    max_seconds: int = 10,
    max_retries: int = 5,
) -> Callable[[Any], Any]:
    # Wait 2^x * 1 second between each retry starting with
    # 4 seconds, then up to 10 seconds, then 10 seconds afterwards
    return retry(
        reraise=True,
        stop=stop_after_attempt(max_retries),
        wait=wait_exponential(multiplier=1, min=min_seconds, max=max_seconds),
        retry=(
            retry_if_exception_type(openai.error.Timeout)
            | retry_if_exception_type(openai.error.APIError)
            | retry_if_exception_type(openai.error.APIConnectionError)
            | retry_if_exception_type(openai.error.RateLimitError)
            | retry_if_exception_type(openai.error.ServiceUnavailableError)
        ),
    )


# Default retry decorator
with_retries = create_retry_decorator()



# flake8: noqa

from .qa import *
from .self_ask import *
from .subqueries import *
from .summarize import *
from .tool import *

__all__ = [
    "qa_bot",
    "self_ask_bot",
    "subqueries_bot",
    "summarize_bot",
    "tool_bot",
]



from typing import (
    List,
)

import pydantic

from ..types.messages import (
    BaseMessageCollection,
    MessageRole,
    Message,
)


__all__ = [
    "System",
    "User",
    "Assistant",
    "ExampleUser",
    "ExampleAssistant",
    "MessageRole",
    "Message",
    "BaseMessageCollection",
    "Example",
]


# Models
class System(Message):
    """System message"""

    role: MessageRole = "system"


class User(Message):
    """User message"""

    role: MessageRole = "user"


class Assistant(Message):
    """Assistant message"""

    role: MessageRole = "assistant"
    forward: bool = True


class ExampleUser(Message):
    """User example message"""

    role: MessageRole = "example_user"


class ExampleAssistant(Message):
    """Assistant example message"""

    role: MessageRole = "example_assistant"


# Abstract implementations
class Example(BaseMessageCollection, pydantic.BaseModel):
    user: str
    assistant: str

    async def get(self) -> List[Message]:
        return [
            ExampleUser(content=self.user),
            ExampleAssistant(content=self.assistant),
        ]



from ..structs import Generate, User
from ..turbo import turbo

__all__ = [
    "qa_bot",
]

# Templates
TEMPLATE = """
Answer the question based on the context below.

START CONTEXT
{{context}}
END CONTEXT

{{question}}
""".strip()


@turbo(temperature=0.1)
async def qa_bot(question: str, context: str):
    """Takes a question and context as input and generate answer based on them"""

    yield User(
        template=TEMPLATE,
        variables=dict(question=question, context=context),
    )

    yield Generate()



# flake8: noqa
# This is so ruff doesn't remove * imports

from jinja2 import Environment
from jinja2schema import infer, to_json_schema
from jsonschema import validate

from .args import get_input_signature
from .lang import inflect

__all__ = [
    "render_template",
]

# jinja environment
jinja_env: Environment = Environment(
    autoescape=False,
    trim_blocks=True,
    lstrip_blocks=True,
)

# Add custom filters
jinja_env.filters["inflect"] = inflect
jinja_env.filters["input_signature"] = get_input_signature


# Funcs
def render_template(template_string: str, variables: dict, check: bool = False) -> str:
    # Parse template
    template = jinja_env.from_string(template_string)

    # If check is required, get required vars from template and validate variables
    if check:
        schema = to_json_schema(infer(template_string))
        validate(instance=variables, schema=schema)

    # Render
    rendered = template.render(**variables)
    return rendered



# flake8: noqa
# This is so ruff doesn't remove * imports

from .local_memory import *
from .summary_memory import *
from .truncated_memory import *

__all__ = [
    "LocalMemory",
    "LocalTruncatedMemory",
    "LocalSummarizeMemory",
]



TOOLBOT_TEMPLATE = """
{%- if prologue -%}
{{prologue}}

---
{% endif %}
You are talking to a {{user_type}}. {{instruction}}

I am simply facilitating the conversation, I will tell you what the {{user_type}} is saying and whenever you are ready to respond to the {{user_type}}, I will relay your response to them. In order to assist the {{user_type}}, you have access to certain tools as specified below. You can get the information you need or take actions on behalf of the {{user_type}} by using the appropriate tools.

I’ll tell you what the {{user_type}} said in this format:
```
{{user_type|capitalize}} said: <what the {{user_type}} said>
```

When you are ready to respond after thinking it through, reply to me in this format:
```
Response: <what you want to say>
```

Keep your responses short and to the point.

Remember that the {{user_type}} ONLY gets to see what you tell me in the format `Response: ...`. As a facilitator, I can strictly only relay the messages between you and the {{user_type}}. Any information you need can only come from either the {{user_type}} answering a follow up question from you or as the result of one of the tools. You can ask the {{user_type}} as many follow up questions as you need.

---

Tools:
{% for tool in tools %}
{% set tool_signature = tool | input_signature -%}
- {{tool.__name__}}: {{tool.__doc__}}. This tool{% if tool_signature %} requires a valid json as input that includes these keys: ({{tool_signature}}){% else %} does not require an input{% endif %}.
{% endfor %}

Think step by step to process the {{user_type}}'s request and use tools to collect required information or take actions. Do NOT make up answers and politely apologize if unable to help even after using the tools.

To use a tool, you must reply to me in exactly the following format:

```
{{user_type | capitalize}} said: <what the {{user_type}} said>
Thought: Need to use a tool? <Yes or No>
Tool: <the name of the tool to use>
Tool Input: <the input to the tool in the format specified by the tool>
```

{% set tool_names = tools|map(attribute="__name__")|list|join(', ') -%}
- Do not add additional spacing or omit any of the fields (Thought, Tool, Tool Input)
- Tool MUST be one of [{{tool_names}}]. No other tools are available.
- Tool Input should be as detailed and specific as possible.

I will give you the result of the tool in this format:
```
Tool Result: <the result of the tool you used>
```

In order to execute complex requests from the user, think step by step and break down the request into a series of sub problems that you can solve by using multiple different tools, one tool at a time one after the other.

After you are done, {{user_type}} only sees what you ask me to relay using `Response: ...`

---

Begin!

{% if additional_info -%}
{{additional_info}}

{% endif %}
{{user_type | capitalize}} said: <what the {{user_type}} said>
""".strip()  # noqa: E501

EXAMPLE_TEMPLATE = """
START EXAMPLE
{{example}}
END EXAMPLE
""".strip()



from datetime import datetime
from functools import wraps
import inspect
from typing import (
    Any,
    Callable,
    Optional,
    Type,
)

from .chat import run_chat
from .config import TurboModel
from .errors import InvalidValueYieldedError
from .memory import LocalMemory

from .structs import (
    Assistant,
    Generate,
    GetInput,
    Result,
    Start,
)

from .structs.proxies import proxy_turbo_gen_fn

from .types import (
    BaseCache,
    BaseMemory,
    BaseMessageCollection,
    Message,
    TurboGenWrapper,
)

from .types.generators import (
    TurboGenFn,
    TurboGenTemplateFn,
)

from .utils.args import ensure_args


__all__ = [
    "turbo",
]


# Decorator
def turbo(
    memory_class: Type[BaseMemory] = LocalMemory,
    model: TurboModel = "gpt-3.5-turbo",
    stream: bool = False,
    cache_class: Optional[Type[BaseCache]] = None,
    debug: Optional[Callable[[dict], None]] = None,
    **kwargs,
) -> Callable[[TurboGenTemplateFn], TurboGenFn]:
    """Parameterized decorator for creating a chatml app from an async generator"""

    # Streaming not supported yet
    if stream:
        raise NotImplementedError("Streaming not supported yet")

    # Settings
    settings = dict(
        memory_class=memory_class,
        model=model,
        stream=stream,
        cache_class=cache_class,
        debug=debug,
    )

    # Prepare openai args
    chat_completion_args = {
        **kwargs,
        "model": model,
        "stream": stream,
    }

    # Parameterized decorator fn
    def wrap_turbo_gen_fn(gen_fn: TurboGenTemplateFn) -> TurboGenFn:
        """Wrapper for chatml app async generator"""

        @wraps(gen_fn)
        async def turbo_gen_fn(**context) -> TurboGenWrapper:
            """Wrapped chatml app from an async generator"""

            # Init memory
            memory = memory_class(model=model)
            assert ensure_args(memory.setup, context)
            await memory.setup(**context)

            # Init cache
            cache = None
            if cache_class:
                cache = cache_class()
                assert ensure_args(cache.setup, context)
                await cache.setup(**context)

            # Init generator
            signature = inspect.signature(gen_fn)
            arg_names = [k for k in signature.parameters.keys()]

            if "memory" in arg_names:
                context["memory"] = memory

            if "cache" in arg_names:
                context["cache"] = cache

            turbo_gen = gen_fn(**context)

            # Parameters
            payload: Any = None

            # Yield Start() to indicate start
            yield Start()

            try:
                while True:
                    # Step through the wrapped generator
                    output = await turbo_gen.asend(payload)

                    # Pass inputs & outputs to debug log if needed
                    if debug:
                        params = dict(
                            app=gen_fn.__name__,
                            timestamp=datetime.utcnow(),
                        )

                        payload and debug(
                            {**params, "type": "input", "payload": payload}
                        )
                        output and debug(
                            {**params, "type": "output", "payload": output}
                        )

                    payload = None

                    # Add to memory
                    if isinstance(output, Result):
                        yield output

                    elif isinstance(output, Message):
                        await memory.append(output)

                        # Yield to user if forward
                        if output.forward:
                            yield Result.from_message(output)

                    elif isinstance(output, BaseMessageCollection):
                        await memory.extend(output)

                    # Yield to user if GetInput
                    elif isinstance(output, GetInput):
                        await memory.append(Assistant(content=output.content))

                        # Get input
                        payload = yield Result.from_message(output)
                        assert payload, f"User input was required, {payload} passed"

                    # Generate result
                    elif isinstance(output, Generate):
                        output_dict = output.dict()
                        forward = output_dict.pop("forward")

                        payload = await run_chat(
                            memory,
                            cache,
                            **{
                                **chat_completion_args,
                                **output_dict,
                            },
                        )

                        # Yield generated result if needed
                        if forward:
                            yield Result.from_message(payload)

                    else:
                        raise InvalidValueYieldedError(
                            f"Invalid value yielded by generator: {type(output)}"
                        )

            except StopAsyncIteration:
                # Generator over, exit
                pass

        # Add reference to the original function
        turbo_gen_fn = proxy_turbo_gen_fn(turbo_gen_fn)
        turbo_gen_fn.fn = gen_fn
        turbo_gen_fn.settings = settings
        turbo_gen_fn.configure = lambda **new_settings: turbo(
            **{**settings, **new_settings}
        )(gen_fn)

        return turbo_gen_fn

    return wrap_turbo_gen_fn



from typing import cast, Generic, List, TypeVar

import dirtyjson
from parse import search as parse_search, with_pattern

__all__ = [
    "Scratchpad",
]

ST = TypeVar("ST")  # , bound=TypedDict) # Not available in py3.8

yesno_mapping = {
    "yes": True,
    "no": False,
    "on": True,
    "off": False,
    "true": True,
    "false": False,
    "1": True,
    "0": False,
}


@with_pattern(r"|".join(yesno_mapping))
def parse_yesno(text: str) -> bool:
    return yesno_mapping[text.lower()]


@with_pattern(r"(\{.+\})")
def parse_json(text: str) -> dict:
    return dirtyjson.loads(text)


class Scratchpad(Generic[ST]):
    def __init__(self, spec: str):
        self.spec = spec

    def _search(self, spec: str, input: str):
        """Use parse.search to parse scratchpad according to spec."""

        return parse_search(
            spec + "\n",
            input + "\n",
            extra_types=dict(bool=parse_yesno, json=parse_json),
        )

    def parse(self, input: str) -> ST:
        """Parse the input string according to the spec."""
        """Only remembers latest results."""

        # Create parsers
        line_specs: List[str] = self.spec.split("\n")

        # Parse the scratchpad
        parsed = [self._search(spec, input) for spec in line_specs]

        # Collect results
        result = {}
        for parsed_line in parsed:
            p = parsed_line.named if parsed_line else {}
            result = {**result, **p}

        return cast(ST, result)



from typing import cast, Union

from .structs import Result, Start

from .errors import GeneratorAlreadyExhausted, GeneratorAlreadyExhaustedError
from .types import TurboGenWrapper

__all__ = [
    "run",
]


async def run(
    gen: TurboGenWrapper,
    input: Union[str, dict],
) -> Result:
    """Run a turbo app"""

    # Set placeholder values
    done = False
    output = GeneratorAlreadyExhausted()

    # Run generator
    try:
        while output := await gen.asend(input):
            if isinstance(output, Start):
                continue

            if output.needs_input:
                break

    # Generator exhausted, mark done
    except StopAsyncIteration:
        done = True

    # Output is still placeholder? Raise error
    if isinstance(output, GeneratorAlreadyExhausted):
        raise GeneratorAlreadyExhaustedError()

    # Set result values
    result = cast(Result, output)
    result.done = done

    return result



from typing import Any, Protocol

import pydantic

from .signals import GetInput

__all__ = [
    "Result",
]


class HasContent(Protocol):
    content: str


class Result(pydantic.BaseModel):
    """Holds the result yielded by a turbo app."""

    content: Any
    needs_input: bool = False
    done: bool = False

    @classmethod
    def from_message(
        cls,
        message: HasContent,
        done: bool = False,
    ) -> "Result":
        assert hasattr(message, "content"), "Content holder object required"

        return cls(
            content=message.content,
            needs_input=isinstance(message, GetInput),
            done=done,
        )



# flake8: noqa
# This is so ruff doesn't remove * imports

from .lang import *
from .retries import *
from .template import *
from .tokens import *

__all__ = [
    "inflect",
    "render_template",
    "create_retry_decorator",
    "with_retries",
    "count_tokens",
    "get_max_tokens_length",
]



from typing import (
    Any,
)

import pydantic


from ..types.cache import BaseCache

__all__ = [
    "SimpleCache",
]


class SimpleCache(BaseCache, pydantic.BaseModel):
    """Simple in-memory cache"""

    cache: dict = {}

    async def has(self, key) -> bool:
        return self.to_key(key) in self.cache

    async def set(self, key, value) -> None:
        self.cache[self.to_key(key)] = self.serialize(value)

    async def get(self, key) -> Any:
        assert await self.has(key), "No cache entry found"
        return self.deserialize(self.cache[self.to_key(key)])

    async def clear(self) -> Any:
        self.cache = {}



# flake8: noqa
# This is so ruff doesn't remove * imports

from .messages import *
from .result import *
from .scratchpad import Scratchpad
from .signals import *

__all__ = [
    "System",
    "User",
    "Assistant",
    "ExampleUser",
    "ExampleAssistant",
    "Generate",
    "GetInput",
    "Example",
    "Scratchpad",
    "Result",
    "Start",
]



from typing import List


from ..structs import User
from ..types.messages import MessageDict
from .local_memory import LocalMemory
from .truncated_memory import MemoryTruncation

__all__ = [
    "LocalSummarizeMemory",
    "MemorySummarization",
]


# Abstract implementations
class MemorySummarization(MemoryTruncation):
    """Mixin for automatic summarization"""

    async def prepare_prompt(
        self,
        max_tokens: int = 0,
    ) -> List[MessageDict]:
        """Summarize the conversation."""

        # To avoid circular imports
        from ..bots import summarize_bot

        # Get first, middle, last message
        messages = await super().prepare_prompt()  # type: ignore
        first, *rest = messages
        if not len(rest):
            return messages

        *middle, last = rest
        if not len(middle):
            return messages

        # Summarize the middle conversation
        conversation = "\n".join(
            [f"{message['role']}: {message['content']}" for message in messages]
        )

        summary = await summarize_bot(text=conversation, text_type="conversation").run()

        summary_as_user = User(summary.content).dict()

        return [first, summary_as_user, last]


class LocalSummarizeMemory(MemorySummarization, LocalMemory):
    """Local memory with automatic summarization"""

    ...



#!/usr/bin/env python3

from typing import List, Literal


__all__ = [
    "TurboModel",
    "available_models",
]

# Allowed chatgpt model names
TurboModel = Literal[
    "gpt-4",
    "gpt-4-0314",
    "gpt-4-32k",
    "gpt-4-32k-0314",
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-0301",
]

available_models: List[TurboModel] = [
    "gpt-4",
    "gpt-4-0314",
    "gpt-4-32k",
    "gpt-4-32k-0314",
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-0301",
]



from typing import List

from ..types.memory import BaseMemory
from ..types.messages import Message

__all__ = [
    "LocalMemory",
]


# Abstract implementations
class LocalMemory(BaseMemory):
    """Store messages in an in-memory list"""

    state: dict = {}
    messages: List[Message] = []

    async def get(self) -> List[Message]:
        return [message for message in self.messages]

    async def extend(self, items) -> None:
        self.messages.extend(items)

    async def get_state(self) -> dict:
        return self.state

    async def set_state(self, new_state: dict, merge: bool = False) -> None:
        self.state = (
            new_state
            if not merge
            else {
                **self.state,
                **new_state,
            }
        )



from typing import Optional, TypedDict

from ...structs import Scratchpad

__all__ = [
    "ParsedQueries",
    "scratchpad",
]


class ParsedQueries(TypedDict):
    query1: Optional[str]
    query2: Optional[str]
    query3: Optional[str]
    query4: Optional[str]
    query5: Optional[str]
    query6: Optional[str]
    query7: Optional[str]
    query8: Optional[str]
    query9: Optional[str]
    query10: Optional[str]


scratchpad: Scratchpad[ParsedQueries] = Scratchpad[ParsedQueries](
    """
1. {query1}
2. {query2}
3. {query3}
4. {query4}
5. {query5}
6. {query6}
7. {query7}
8. {query8}
9. {query9}
10. {query10}
""".strip()
)



from typing import List, Optional

from ...structs import Assistant, GetInput, Generate, Result, User
from ...turbo import turbo
from ...types import Tool
from .scratchpad import scratchpad
from .template import EXAMPLE_TEMPLATE, TOOLBOT_TEMPLATE

default_tool_example: str = """
Customer said: Where is the store located?
Thought: Use another tool to get information? Yes
Tool: GetInformation
Tool Input: What is the address of the store?
Tool Result: The address of [STORE NAME] is [SOME ADDRESS]
Thought: Use another tool to get information? No
Response: [STORE NAME] is located at [SOME ADDRESS].
""".strip()


@turbo(model="gpt-3.5-turbo", temperature=0.7)
async def tool_bot(
    tools: List[Tool],
    prologue: Optional[str] = None,
    user_type: str = "user",
    instruction: Optional[str] = None,
    example: str = default_tool_example,
    max_iterations: int = 6,
    # Placeholder for message to send to the user
    message: str = "How can I help you?",
):
    # Yield instructions
    yield User(
        template=TOOLBOT_TEMPLATE,
        variables=dict(
            tools=tools,
            prologue=prologue,
            user_type=user_type,
            instruction=instruction,
            additional_info=None,
        ),
        # check=True,  # FIXME: Throws an error, possible bug in jinja2schema
    )

    # Yield example
    yield User(template=EXAMPLE_TEMPLATE, variables={"example": example})

    # Get user input
    input = yield GetInput(content=message)
    yield User(content=f"{user_type} said: {input}")

    # Start the tool agent
    parsed_tools = {}
    iterations_left = max_iterations
    tools_used = []

    # Keep running until we get final response
    while "final_response" not in parsed_tools and iterations_left > 0:
        iterations_left -= 1

        output = yield Generate(stop=["Tool Result"], forward=False)
        parsed_tools = scratchpad.parse(output.content)

        # If no tool required to run, continue
        if not parsed_tools.get("tool_name"):
            continue

        # Otherwise run tool
        # First check if tool is valid
        selected_tool = parsed_tools["tool_name"]
        tool_names = [tool.__name__ for tool in tools]

        if selected_tool not in tool_names:
            yield User(
                content=(
                    f"Tool Result: `{selected_tool}` is not a valid tool."
                    f"Pick one from [{', '.join(tool_names)}]."
                )
            )

            continue

        # Tool is valid. Run it
        tools_used.append(selected_tool)

        # Get tool input
        tool_fn = next(tool for tool in tools if tool.__name__ == selected_tool)
        tool_input = parsed_tools.get("tool_input") or {}

        # Run tool
        tool_result = await tool_fn(**tool_input)

        # Yield the observation
        yield Assistant(content=f"Tool Result: {tool_result.strip()}")

    # Yield final response if any
    response = parsed_tools.get(
        "final_response",
        "I am not sure how to answer this question",
    )

    yield Result(content=dict(response=response, tools_used=tools_used))



from typing import Callable, Set

import inspect
from inspect import Parameter as P


def get_input_signature(fn: Callable) -> str:
    return ", ".join([str(p) for p in inspect.signature(fn).parameters.values()])


def get_required_args(fn: Callable) -> Set[str]:
    """Get required args for a function."""

    sig = inspect.signature(fn)
    params = list(sig.parameters.values())

    valid_param_kinds = [
        P.POSITIONAL_ONLY,
        P.POSITIONAL_OR_KEYWORD,
        P.KEYWORD_ONLY,
    ]

    required_params = [
        p.name for p in params if (p.default is p.empty and p.kind in valid_param_kinds)
    ]

    return set(required_params)


def ensure_args(fn: Callable, args: dict) -> bool:
    """Ensure that args dict has all required args for fn."""

    required_args: Set[str] = get_required_args(fn)
    required_intersection = required_args.intersection(set(args.keys()))
    has_all_required = required_intersection == required_args

    return has_all_required



from typing import Dict, List

import tiktoken

from ..config import TurboModel

__all__ = [
    "count_tokens",
    "get_max_tokens_length",
]

# See: https://platform.openai.com/docs/models/gpt-4
MODEL_WINDOWS: Dict[str, int] = {
    "gpt-4": 8_192,
    "gpt-4-32k": 32_768,
    "gpt-3.5-turbo": 4_096,
    "text-davinci": 4_097,
    "text-curie": 2_049,
    "text-babbage": 2_049,
    "text-ada": 2_049,
    "code-davinci": 8_001,
    "code-cushman": 2_048,
    "davinci": 2_049,
    "curie": 2_049,
    "babbage": 2_049,
    "ada": 2_049,
}


def get_max_tokens_length(model_name: str) -> int:
    for model_prefix, window in MODEL_WINDOWS.items():
        if model_name.startswith(model_prefix):
            return window

    raise ValueError(f"{model_name} is not a known openai model")


def count_tokens(messages: List[dict], model: TurboModel) -> int:
    """Count the number of tokens stored in list of messages."""

    # Get tiktoken encoding
    encoding = tiktoken.encoding_for_model(model)

    ###########
    # From https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb  # noqa: E501
    ###########
    if model == "gpt-3.5-turbo-0301":  # note: future models may deviate from this
        num_tokens = 0
        for message in messages:
            # every message follows <im_start>{role/name}\n{content}<im_end>\n
            num_tokens += 4
            for key, value in message.items():
                num_tokens += len(encoding.encode(value))
                if key == "name":  # if there's a name, the role is omitted
                    num_tokens += -1  # role is always required and always 1 token
        num_tokens += 2  # every reply is primed with <im_start>assistant
        return num_tokens
    ###########

    # Else, just sum up
    texts: List[str] = [message["content"] for message in messages]
    tokens_list: List[List[int]] = [encoding.encode(text) for text in texts]

    count: int = 2 + sum([len(tokens) + 4 for tokens in tokens_list])

    return count



from typing import (
    Any,
    AsyncGenerator,
    Optional,
    Protocol,
    Union,
)


from ..structs.proxies import TurboGenWrapper
from ..structs.result import Result
from ..structs.signals import Generate, GetInput

from .messages import Message
from .memory import BaseMemory

__all__ = [
    "TurboGenWrapper",
]


# Types
TurboGenTemplate = AsyncGenerator[
    Union[Message, Generate, GetInput, Result],
    Any,
]


class TurboGenTemplateFn(Protocol):
    def __call__(
        self,
        memory: Optional[BaseMemory] = None,
        **context,
    ) -> TurboGenTemplate:
        ...


class TurboGenFn(Protocol):
    fn: TurboGenTemplateFn
    settings: dict

    def configure(self, new_settings: dict) -> "TurboGenFn":
        ...

    def __call__(
        self,
        **context,
    ) -> TurboGenWrapper:
        ...



import pydantic


__all__ = [
    "Generate",
    "GetInput",
    "Start",
]


# Signal classes
class Generate(pydantic.BaseModel, extra=pydantic.Extra.allow):
    """Placeholder value to indicate that completion should be run"""

    forward: bool = True


class GetInput(pydantic.BaseModel):
    """Placeholder value to indicate that user input is needed"""

    content: str = "User input needed"


class Start:
    """Placeholder value to indicate generator start"""



from typing import List, Tuple

from ..structs import Result
from ..turbo import turbo
from ..types.generators import TurboGenFn
from .subqueries import subqueries_bot

__all__ = [
    "self_ask_bot",
]


# Utils
def make_qa_context(qa_context, previous_qa):
    """Prepare qa context."""

    # Add previous q&a as faq
    if len(previous_qa):
        qas = "\n\n".join([f"Q: {q}\nA: {a}" for q, a in previous_qa])

        qa_context += "FAQ:\n"
        qa_context += qas

    return qa_context


@turbo(temperature=0.7)
async def self_ask_bot(
    question: str,
    context: str,
    qa_bot: TurboGenFn,
    subquery_instructions: str = "User is asking questions to an AI assistant.",
):
    """Takes a question and qa_bot and uses them to answer step by step."""

    # Generate sub queries
    queries = await subqueries_bot(
        request=question,
        context=subquery_instructions,
    ).run()

    # Answer sub-questions
    previous_qa: List[Tuple[str, str]] = []

    for query in [*queries.content, question]:
        # Generate new answer
        answer = await qa_bot(
            question=query,
            context=make_qa_context(context, previous_qa),
        ).run()

        # Append query and answer to list
        previous_qa.append((query, answer.content))

    # Yield the last answer as the result
    yield Result(content=previous_qa[-1][-1])



from abc import ABC, abstractmethod
import json
from typing import (
    Any,
)

from .misc import WithSetup


__all__ = [
    "BaseCache",
]


# Abstract classes
class BaseCache(ABC, WithSetup):
    """Base class for caching agent responses"""

    def serialize(self, obj: Any) -> Any:
        return obj

    def deserialize(self, hashed: Any) -> Any:
        return hashed

    def to_key(self, obj: Any) -> str:
        return json.dumps(self.serialize(obj))

    @abstractmethod
    async def has(self, key: Any) -> bool:
        ...

    @abstractmethod
    async def set(self, key: Any, value: Any) -> None:
        ...

    @abstractmethod
    async def get(self, key: Any) -> Any:
        ...

    @abstractmethod
    async def clear(self) -> Any:
        ...



SUBQUERIES_TEMPLATE = """
{% set _request_type = request_type or "question" -%}
{% set _solve_act = solve_act or "answer" -%}
{% set _max_queries = max_queries or 6 -%}
Your task is to create a plan to collect information for {{_solve_act | inflect("VBG")}} a complex {{_request_type}}.
You have at your disposal a knowledgebase that can give answers to plain english queries.
Based on the given {{_request_type}} and the context below, write a list of simple, standalone queries to ask the knowledgebase in order to collect all the information necessary to {{_solve_act}} that {{_request_type}}.
You do not have to {{_solve_act}} the {{_request_type}}, just come up with a step by step plan to collect information to do so.

Instructions:
1. First think step by step to break down the {{_request_type}} into individual queries.
2. Use the format below to think through and write your thoughts.
3. Write down the knwoledgebase queries as a numbered list under the heading: "Queries to execute:"
4. Each query MUST be concise and standalone.
5. Do not come up with more than {{_max_queries}} queries.
6. The queries will be executed verbatim so make sure to resolve all coreferences and add contextual information.

Format for writing down your thoughts and the queries:

```
Thoughts:
- The {{_request_type}} is about "<topic>"
- The {{_request_type}} requires looking up ...
- ... other thoughts
- In order to {{_solve_act}} "<{{_request_type}}>" for the user, we need to first ask the knowledgebase the following queries.

Queries to execute:
1. <query 1>
2. <query 3>
and so on...

For example,

START EXAMPLE
{{example}}
END EXAMPLE

---

Begin!

Context:
{{context}}

{{_request_type | capitalize}}:
{{request}}
```
""".strip()  # noqa: E501



from typing import Optional, Union

from peak.util.proxies import ObjectWrapper

from .signals import Start


class TurboGenWrapper(ObjectWrapper):
    g = None

    def __init__(self, g):
        ObjectWrapper.__init__(self, g)
        self.g = g

    def __aiter__(self):
        return self.g.__aiter__()

    async def init(self):
        # Send none to init generator
        start_value = await self.g.asend(None)
        assert isinstance(start_value, Start), "Called .init(ß) after start"

        # Enable fluent chaining
        return self

    async def run(
        self,
        input: Optional[Union[str, dict]] = None,
    ):
        # To avoid circular import
        from ..runner import run as gen_run

        return await gen_run(self, input)


def proxy_turbo_gen_fn(turbo_gen_fn):
    def wrapped_fn(*args, **kwargs):
        gen = turbo_gen_fn(*args, **kwargs)
        proxy = TurboGenWrapper(gen)

        return proxy

    return wrapped_fn



from typing import (
    Optional,
)

import openai


from .structs import Assistant
from .types import (
    BaseCache,
    BaseMemory,
)


from .utils import with_retries

__all__ = [
    "run_chat",
]


# Chat runner
@with_retries
async def run_chat(
    memory: BaseMemory,
    cache: Optional[BaseCache] = None,
    **kwargs,
) -> Assistant:
    """Run ChatCompletion for the memory so far"""

    # Get messages from memory
    max_tokens = kwargs.get("max_tokens")
    prompt = await memory.prepare_prompt(max_tokens=max_tokens)

    # Check cache
    if cache and await cache.has(prompt):
        cached = await cache.get(prompt)
        return Assistant(**cached)

    # Create completion
    chat_completion = await openai.ChatCompletion.acreate(
        messages=prompt,
        **kwargs,
    )

    # Parse result
    output = chat_completion.choices[0].message
    payload = dict(content=output["content"])
    result = Assistant(**payload)

    # Append result to memory
    await memory.append(result)

    # Add to cache
    if cache:
        await cache.set(prompt, payload)

    return result



from abc import abstractmethod
from itertools import groupby
from operator import attrgetter
from typing import cast, List

import pydantic

from ..config import TurboModel
from .messages import (
    BaseMessageCollection,
    MessageDict,
    Message,
)

from .misc import WithSetup


__all__ = [
    "BaseMemory",
]


# Abstract classes
class BaseMemory(BaseMessageCollection, WithSetup, pydantic.BaseModel):
    """Base class for persisting conversation history and state."""

    model: TurboModel

    @abstractmethod
    async def extend(self, items: List[Message]) -> None:
        ...

    @abstractmethod
    async def get_state(self) -> dict:
        ...

    @abstractmethod
    async def set_state(self, new_state: dict, merge: bool = False) -> None:
        ...

    async def append(self, item: Message) -> None:
        await self.extend([item])

    async def prepare_prompt(
        self,
        max_tokens: int = 0,
    ) -> List[MessageDict]:
        """Turn message history into a prompt for openai."""

        messages = await self.get()
        messages = sorted(messages, key=attrgetter("label", "timestamp"))

        sticky_top, sticky_bottom, non_sticky = (
            [
                message
                for message in messages
                if message.sticky and message.sticky_position == "top"
            ],
            [
                message
                for message in messages
                if message.sticky and message.sticky_position == "bottom"
            ],
            [message for message in messages if not message.sticky],
        )

        # Take only last for every label group
        sticky_top = [
            list(group)[-1] for _, group in groupby(sticky_top, key=attrgetter("label"))
        ]

        sticky_bottom = [
            list(group)[-1]
            for _, group in groupby(sticky_bottom, key=attrgetter("label"))
        ]

        messages = sticky_top + non_sticky + sticky_bottom

        # Convert
        message_dicts = [cast(MessageDict, message.dict()) for message in messages]

        # Add prefix (system name=) for examples
        for i, message in enumerate(message_dicts):
            if message["role"].startswith("example"):
                message_dicts[i]["role"] = f"system name={message['role']}"

        # max_tokens will be used by extending classes

        return message_dicts



# flake8: noqa
# This is so ruff doesn't remove * imports

from .cache import *
from .generators import *
from .memory import *
from .messages import *
from .tools import Tool

__all__ = [
    "MessageRole",
    "Message",
    "MessageDict",
    "BaseMessageCollection",
    "TurboGenWrapper",
    "BaseMemory",
    "BaseCache",
    "Tool",
]



import pydantic

__all__ = [
    "InvalidValueYieldedError",
    "GeneratorAlreadyExhaustedError",
]


# Models
class InvalidValueYieldedError(ValueError):
    """Invalid value yielded by generator"""

    ...


class GeneratorAlreadyExhaustedError(StopAsyncIteration):
    """Generator was already exhausted"""

    ...


# Placeholder classes
class GeneratorAlreadyExhausted(pydantic.BaseModel):
    """Placeholder value to indicate that the generator was already exhausted"""

    ...



from typing import List


from ..config import TurboModel
from ..types.messages import MessageDict
from ..utils import count_tokens, get_max_tokens_length
from .local_memory import LocalMemory

__all__ = [
    "LocalTruncatedMemory",
    "MemoryTruncation",
]


# Abstract implementations
class MemoryTruncation:
    """Mixin for automatic truncation"""

    model: TurboModel

    async def prepare_prompt(
        self,
        max_tokens: int = 0,
    ) -> List[MessageDict]:
        """Drop messages iteratively to keep max_tokens + length < context_window."""
        """Keep first and last message."""

        max_tokens = max_tokens or 0
        context_window: int = get_max_tokens_length(self.model) - max_tokens
        messages = await super().prepare_prompt()  # type: ignore

        # Make sure first message is within limits
        first, *rest = messages
        assert count_tokens([first], self.model) <= context_window

        if not len(rest):
            return messages

        # Make sure first and last combined are also within limits
        *middle, last = rest
        total = count_tokens([first, last], self.model)
        assert total <= context_window

        # Iteratively filter middle (remove from end)
        final = [first]

        while (
            len(middle)
            and (total := total + count_tokens(middle[-1:], self.model))
            <= context_window
        ):
            final.insert(1, middle.pop())

        return final


class LocalTruncatedMemory(MemoryTruncation, LocalMemory):
    """Local memory with automatic truncation"""

    ...



from typing import Optional, TypedDict

from ...structs import Scratchpad

__all__ = [
    "ParsedTools",
    "scratchpad",
]


class ParsedTools(TypedDict):
    should_use_tool: Optional[bool]
    tool_name: Optional[str]
    tool_input: Optional[dict]
    final_response: Optional[str]


scratchpad: Scratchpad[ParsedTools] = Scratchpad[ParsedTools](
    """
Tool: {tool_name:S}
Tool Input: {tool_input:json}
Response: {final_response}
""".strip()
)



from typing import List

from ...structs import Generate, Result, User
from ...turbo import turbo
from .scratchpad import scratchpad
from .template import SUBQUERIES_TEMPLATE

__all__ = [
    "subqueries_bot",
]

default_subqueries_example: str = """
Context:
User is a customer at a grocery store and is asking the question to the store manager.

Question:
What is the least expensive cereal that is healthy and has a low calorie content but is also tasty?

Thoughts:
- The user wants to know about "tasty but healthy cereal".
- The question requires looking up cereal by nutrition value and price
- In order to "What is the least expensive cereal that is healthy and has a low calorie content but is also tasty?" for the user, we need to first ask the knowledgebase the following queries:

Queries to execute:
1. What are some tasty cereal that are healthy?
2. What are the prices of the above cereals?
3. What is the least expensive cereal of the above?
""".strip()  # noqa: E501


@turbo(model="gpt-4")
async def subqueries_bot(
    request: str,
    context: str,
    example: str = default_subqueries_example,
    max_queries: int = 6,
    request_type: str = "question",
    solve_act: str = "answer",
):
    """Bot that decomposes a request into sub queries for a knowledgebase."""

    # Yield instructions
    yield User(
        template=SUBQUERIES_TEMPLATE,
        variables=dict(
            request=request,
            context=context,
            example=example,
            max_queries=max_queries,
            request_type=request_type,
            solve_act=solve_act,
        ),
    )

    # Generate output
    output = yield Generate(forward=False)

    # Parse output
    parsed_queries = scratchpad.parse(output.content)
    queries: List[str] = [
        query for query in parsed_queries.values() if query is not None
    ]

    yield Result(content=queries)



from typing import Any, List


class PickOptionalOff:
    pass


def pick(
    dictionary: dict,
    keys: List[str],
    optional: Any = PickOptionalOff(),
) -> dict:
    """Pick keys from dictionary and return as new dict."""

    return {
        key: (
            dictionary.get(key, optional)
            if isinstance(optional, PickOptionalOff)
            else dictionary[key]
        )
        for key in keys
    }



from abc import ABC, abstractmethod
from datetime import datetime
import json
from typing import (
    cast,
    List,
    Literal,
    Optional,
    TypedDict,
    Union,
)

import pydantic

from ..utils import render_template

__all__ = [
    "MessageRole",
    "Message",
    "MessageDict",
    "BaseMessageCollection",
]


# Enums
# Allowed values for openai chatml prefixes
MessageRole = Literal[
    "system",
    "user",
    "assistant",
    "example_user",  # Will add prefix (system name=) in prepare_prompt
    "example_assistant",
]


# Models
class MessageDict(TypedDict):
    role: str
    content: str


class Message(pydantic.BaseModel):
    """Container for a single chatml prefix message"""

    role: MessageRole
    timestamp: datetime = pydantic.Field(default_factory=datetime.utcnow)

    # Content / Template
    content: Optional[Union[str, dict]] = None
    template: Optional[str] = None
    variables: Optional[dict] = None

    # Check template variables
    check: bool = False

    # Stickiness
    sticky: bool = False
    label: str = ""
    sticky_position: Literal["top", "bottom"] = "bottom"

    # Should forward downstream
    forward: bool = False

    @pydantic.root_validator
    def validate_content_template(cls, values: dict) -> dict:
        # Get vals
        content = values["content"]
        check = values["check"]
        template_string = values.pop("template")
        variables = values.pop("variables")
        sticky = values["sticky"]
        label = values["label"]

        # Check that sticky messages have an id
        assert not (sticky ^ bool(label)), "Sticky must be set with label"

        # Check that correct values set
        template_set = bool(template_string) and (variables is not None)
        assert template_set ^ (
            content is not None
        ), "Either content or template/variables must be set"

        # Render template
        if template_set:
            values["content"] = render_template(template_string, variables, check)

        return values

    def dict(self, **kwargs) -> MessageDict:
        # Include role and content
        kwargs.setdefault("include", {"role", "content"})
        data = super().dict(**kwargs)

        # Convert content to json if not str
        if not isinstance(data["content"], str):
            data["content"] = json.dumps(data["content"])

        return cast(MessageDict, data)


# Abstract classes
class BaseMessageCollection(ABC):
    """Base class for async collections of prefix messages"""

    @abstractmethod
    async def get(self) -> List[Message]:
        ...

    async def get_dicts(self) -> List[MessageDict]:
        messages = await self.get()

        # Convert
        message_dicts = [cast(MessageDict, message.dict()) for message in messages]

        return message_dicts



# flake8: noqa

from .bots import *
from .cache import *
from .chat import *
from .config import *
from .errors import *
from .memory import *
from .structs import *
from .turbo import *
from .types import *
from .utils import *

__all__ = [
    "System",
    "User",
    "Assistant",
    "ExampleUser",
    "ExampleAssistant",
    "Generate",
    "GetInput",
    "InvalidValueYieldedError",
    "GeneratorAlreadyExhaustedError",
    "TurboGenWrapper",
    "MessageRole",
    "Message",
    "BaseMessageCollection",
    "BaseMemory",
    "LocalMemory",
    "LocalSummarizeMemory",
    "LocalTruncatedMemory",
    "Example",
    "BaseCache",
    "SimpleCache",
    "Scratchpad",
    "Result",
    "Tool",
    "turbo",
    "summarize_bot",
    "subqueries_bot",
    "tool_bot",
    "self_ask_bot",
    "qa_bot",
    "available_models",
]
